import numpy as np
from scipy.fft import fft, fftfreq
from scipy import signal
from .numerics import decimate

def power_transfer(inputs, outputs, step, **options):
    """
    Power spectrum transfer function from input and output data.

    :param inputs:      input time history.
    :type inputs:       1D array
    :param outputs:     output response history.
    :type outputs:      1D array
    :param step:        timestep.
    :type step:         float
    :param decimation:  decimation factor. default: 1
    :type decimation:   int, optional

    :return:            (periods, amplitudes)
    :rtype:             tuple of arrays
    """
    decimation = options.get("decimation", None)

    if decimation is not None:
        inputs = decimate(inputs, decimation=decimation)
        outputs = decimate(outputs, decimation=decimation)
        step = step*decimation

    assert len(inputs) == len(outputs)
    input_transform = power_spectrum(inputs, step, **options)
    output_transform = power_spectrum(outputs, step, **options)
    return (input_transform[0], output_transform[1]/input_transform[1])

def response_transfer(inputs, outputs, step, **options):
    """
    Response spectrum transfer function from input and output data.
    
    :param inputs:      input time history.
    :type inputs:       1D array
    :param outputs:     output response history.
    :type outputs:      1D array
    :param step:        timestep.
    :type step:         float
    :param pseudo:      if True, uses pseudo accelerations. default: False
    :type pseudo:       bool, optional
    :param decimation:  decimation factor. default: 1
    :type decimation:   int, optional

    :return:            (periods, amplitudes)
    :rtype:             tuple of arrays
    """
    pseudo = options.get("pseudo", False)
    decimation = options.get("decimation", None)

    if decimation is not None:
        inputs = decimate(inputs, decimation=decimation)
        outputs = decimate(outputs, decimation=decimation)
        step = step*decimation

    from sdof import spectrum
    Din,  _,  Ain = spectrum(inputs,  step, **options)
    Dout, _, Aout = spectrum(outputs, step, **options)
    periods = Din[0]

    if pseudo:
        input_spectrum = Din[1,:]*(2*np.pi/periods)**2
    else:
        input_spectrum = Ain[1]

    if pseudo:
        output_spectrum = Dout[1,:]*(2*np.pi/periods)**2
    else:
        output_spectrum = Aout[1]

    return (periods, output_spectrum/input_spectrum)

def fourier_transfer(inputs, outputs, step, **options):
    """
    Fourier spectrum transfer function from input and output data.

    :param inputs:      input time history.
    :type inputs:       1D array
    :param outputs:     output response history.
    :type outputs:      1D array
    :param step:        timestep.
    :type step:         float
    :param decimation:  decimation factor. default: 1
    :type decimation:   int, optional

    :return:            (periods, amplitudes)
    :rtype:             tuple of arrays
    """
    decimation = options.get("decimation", None)

    if decimation is not None:
        inputs = decimate(inputs, decimation=decimation)
        outputs = decimate(outputs, decimation=decimation)
        step = step*decimation

    assert len(inputs) == len(outputs)
    input_transform = fourier_spectrum(inputs, step, **options)
    output_transform = fourier_spectrum(outputs, step, **options)
    return (input_transform[0], output_transform[1]/input_transform[1])

def power_spectrum(series, step, period_band=None, **options):
    """
    Power spectrum of a signal, as a function of period (i.e., periodogram).

    :param series:      time series.
    :type series:       1D array
    :param step:        timestep.
    :type step:         float
    :param period_band: minimum and maximum period of interest, in seconds.
    :type period_band:  tuple, optional

    :return:            (periods, amplitudes)
    :rtype:             tuple of arrays.
    """
    periods, amplitudes = fourier_spectrum(series, step, **options)
    if period_band is not None:
        period_indices = np.logical_and(periods>period_band[0], periods<period_band[1])
        periods = periods[period_indices]
        amplitudes = amplitudes[period_indices]
    return np.array([periods, (np.abs(amplitudes))**2])

# def _power_spectrum(series, step, **options):  # equivalent to power_spectrum()
#     frequencies, power_spectral_density = signal.periodogram(series, step)
#     return (1/frequencies, power_spectral_density)

def fourier_spectrum(series, step, period_band=None, **options):
    """
    Fourier amplitude spectrum of a signal, as a function of period.

    :param series:      time series.
    :type series:       1D array
    :param step:        timestep.
    :type step:         float
    :param period_band: minimum and maximum period of interest, in seconds.
    :type period_band:  tuple, optional

    :return:            (periods, amplitudes)
    :rtype:             tuple of arrays.
    """
    assert len(series.shape) == 1
    N = len(series)
    frequencies = fftfreq(N,step)[1:N//2+2]
    amplitudes = 2.0/N*np.abs(fft(series)[1:N//2+2])
    periods = 1/frequencies
    if period_band is not None:
        period_indices = np.logical_and(periods>period_band[0], periods<period_band[1])
        periods = periods[period_indices]
        amplitudes = amplitudes[period_indices]
    return np.array([periods, amplitudes])

def _newmark():
    pass
 


from mdof import markov, realize, modal
from .numerics import decimate

def system(inputs, outputs, **options):
    """
    State space system realization from input and output data.

    :param inputs:      input time history. dimensions: :math:`(q,nt)`, where
                        :math:`q` = number of inputs, and :math:`nt` = number of timesteps
    :type inputs:       array
    :param outputs:     output response history.
                        dimensions: :math:`(p,nt)`, where :math:`p` = number of outputs, and
                        :math:`nt` = number of timesteps
    :type outputs:      array
    :param method:      system identification method. default is "srim", other options are "okid-era" and "okid-era-dc".
    :type method:       string, optional
    :param decimation:  decimation factor. default: 8
    :type decimation:   int, optional

    :return:            system realization in the form of state space coefficients ``(A,B,C,D)``
    :rtype:             tuple of arrays
    """
    method = options.get("method", "srim")
    decimation = options.get("decimation", None)

    if method not in {
        "srim",
        "okid-era",
        "okid-era-dc"
    }: raise ValueError(f"Unknown method {method}")

    if decimation is not None:
        inputs = decimate(inputs, decimation=decimation)
        outputs = decimate(outputs, decimation=decimation)

    if method == "okid-era":
        Y = markov.okid(inputs, outputs, **options)
        realization = realize.era(Y, **options)

    if method == "okid-era-dc":
        Y = markov.okid(inputs, outputs, **options)
        realization = realize.era_dc(Y, **options)
    
    if method == "srim":
        realization = realize.srim(inputs, outputs, **options)

    return realization


"""
This module implements functions that extract modal information
from a state space realization or transfer function.
"""
import numpy as np
import scipy.linalg as sl
from numpy import pi
from mdof.validation import OutputEMAC, MPC

def _condeig(a): # TODO: make this match matlab source code for condeig
    """
    vals, vecs, conds = condeig(A) Computes condition numbers for the
    eigenvalues of a matrix. The condition numbers are the reciprocals
    of the cosines of the angles between the left and right eigenvectors.
    Inspired by Arno Onken's Octave code for condeig.

    https://github.com/macd/rogues/blob/master/rogues/utils/condeig.py
    """
    m, n = a.shape
    # eigenvalues, left and right eigenvectors
    lamr, vl, vr = sl.eig(a, left=True, right=True)
    vl = vl.T.conj()
    # Normalize vectors
    for i in range(n):
        vl[i, :] = vl[i, :] / np.sqrt(abs(vl[i, :] ** 2).sum())
    # Condition numbers are reciprocal of the cosines (dot products) of the
    # left eignevectors with the right eigenvectors.
    c = abs(1 / np.diag(np.dot(vl, vr))) 
    return vr, lamr, c

def system_modes(realization, dt, **options):
    """
    Modal identification from a state space system realization.

    :param realization:     realization in the form of state space coefficients ``(A,B,C,D)``
    :type realization:      tuple
    :param dt:              timestep.
    :type dt:               float
    :param decimation:      decimation factor. default: 1
    :type decimation:       int, optional
    :param Observability:   Observability matrix; can be reused from :func:`mdof.realize.srim`.
                            default: None
    :type Observability:    array, optional

    :return:                system modes, including natural frequencies, damping ratios, mode shapes,
                            condition numbers, and modal validation metrics EMAC and MPC.
    :rtype:                 dictionary
    """
    decimation = options.get("decimation",
                             1)
    Observability = options.get("Observability",
                                None)
    
    dt = dt*decimation

    A,_,C,_ = realization
    # eigendecomp A
    Psi,Gam,cnd = _condeig(A)  # eigenvectors (Psi) & eigenvalues (Gam) of the matrix A

    # get damping and frequencies from eigendecomp of A
    Lam = (np.log(Gam))/dt
    # TODO: maybe clean
    Omega = np.real((Lam*np.conj(Lam))**0.5)  # radians per second. taking the real part because numpy keeps a +0j.
    # Omega = np.real_if_close((Lam*np.conj(Lam))**0.5) # use real_if_close
    freq = Omega/(2*pi) # cycles per second (Hz)
    Omega.__str__() # helps to avoid "divide by zero" numpy warning message
    # Lam_real = np.real(Lam)
    # Lam_real+=1
    # Lam_real-=1
    # Omega_real = np.real(Omega)
    # Omega_real+=1
    # Omega_real-=1
    # del Omega
    # del Lam
    # import gc
    # gc.collect()
    # print(Omega_real)
    # print(Lam_real)
    # damp = -Lam_real/Omega_real
    damp = -np.real(Lam)/Omega

    # get modeshapes from C and eigendecomp of A
    modeshape = C@Psi

    # energy condensed output EMAC (extended modal amplitude coherence)
    energy_condensed_emaco = OutputEMAC(A,C,Observability=Observability,Psi=Psi,Gam=Gam,**options)

    # MPC (modal phase collinearity)
    mpc = MPC(A,C,Psi=Psi)

    # weed out unique roots: get indices of (1) roots that only show up once, and
    # (2) the first of each pair.
    _, notroots = np.unique(freq.round(decimals=5), return_index=True)
    
    # print(notroots)
    modes = {str(i):
                {'freq': freq[i],  # identified frequency
                'damp': damp[i],   # identified damping ratio
                'modeshape': modeshape[:,i],  # identified modeshape
                'cnd': cnd[i],     # condition number of the eigenvalue
                'energy_condensed_emaco': energy_condensed_emaco[i],  # energy condensed output emac
                'mpc': mpc[i],     # MPC
                }
            for i in range(len(freq)) if i not in notroots # and energy_condensed_emaco[i] > 0.5 and mpc[i] > 0.5
            }

    return modes

def spectrum_modes(periods, amplitudes, **options):
    """
    Modal identification from a transfer function.

    :param periods:     transfer function periods
    :type periods:      array
    :param amplitudes:  transfer function amplitudes
    :type amplitudes:   array

    :return:            (fundamental_periods, fundamental_amplitudes)
    :rtype:             tuple
    """
    from scipy.signal import find_peaks
    height = options.get("height", 0.4)
    width = options.get("width", 0.2)
    rel_height = options.get("rel_height", 0.1)
    
    peaks, _ = find_peaks(amplitudes, height=height, width=width, rel_height=rel_height)
    fundamental_periods = periods[peaks]
    fundamental_amplitudes = amplitudes[peaks]
    
    return (fundamental_periods, fundamental_amplitudes)


import numpy as np
lin_solve = np.linalg.solve
import multiprocessing
from functools import partial
import warnings
try:
    from tqdm import tqdm as progress_bar

except:
    def progress_bar(arg, **kwds): return arg

from . import numerics


def era(Y,**options):
    """
    System realization from Markov parameters (discrete impulse response data).
    Ho-Kalman / Eigensystem Realization Algorithm (ERA) [1]_ [2]_.

    :param Y:       Markov parameters. dimensions: :math:`(p,q,nt)`, where :math:`p` = number of outputs,
                    :math:`q` = number of inputs, and :math:`nt` = number of Markov parameters.
    :type Y:        array
    :param horizon: number of block rows in Hankel matrix = order of observability matrix.
                    default: :math:`min(150, (nt-1)/2)`
    :type horizon:  int, optional
    :param nc:      number of block columns in Hankel matrix = order of controllability matrix.
                    default: :math:`min(150, max(nt-1-` ``horizon``:math:`, (nt-1)/2))`
    :type nc:       int, optional
    :param order:   model order. default: :math:`min(20,` ``horizon``:math:`/2)`
    :type order:    int, optional

    :return:        realization in the form of state space coefficients ``(A,B,C,D)``
    :rtype:         tuple of arrays

    References
    ----------
    .. [1]  Ho, Β. L., & Kálmán, R. E. (1966). Effective construction of linear state-variable models
            from input/output functions: Die Konstruktion von linearen Modeilen in der Darstellung
            durch Zustandsvariable aus den Beziehungen für Ein-und Ausgangsgrößen. at-Automatisierungstechnik,
            14(1-12), 545-548. (https://doi.org/10.1524/auto.1966.14.112.545)
    .. [2]  Juang, J. N., & Pappa, R. S. (1985). An eigensystem realization algorithm for modal parameter
            identification and model reduction. Journal of guidance, control, and dynamics, 8(5), 620-627.
            (https://doi.org/10.2514/3.20031)
    """
    p,q,nt = Y.shape # p = number of outputs, q = number of inputs, nt = number of timesteps

    no = options.get("no",
         options.get("horizon",
                     None))
    nc = options.get("nc",
                     None)

    # get D from first p x q block of impulse response
    Dr = Y[:,:,0]  # first block of output data

    # size of Hankel matrix
    if no is None:
        no = min(150, int((nt-1)/2))
    if nc is None:
        nc = min(150, max(nt-1-no, int((nt-1)/2)))
    # make sure there are enough timesteps to assemble this size of Hankel matrix
    assert nt >= no+nc

    r = options.get("r", 
        options.get("order",
                    min(20, int(no/2))))

    # make impulse response into Hankel matrix and shifted Hankel matrix
    H = np.zeros((p*(no), q*(nc+1)))
    for i in range(no):
        for j in range(nc+1):
            H[p*i:p*(i+1), q*j:q*(j+1)] = Y[:,:,i+j+1]
    H0 = H[:,:-q]
    H1 = H[:,q:]
    assert H0.shape == H1.shape == (p*(no), q*(nc))

    # reduced SVD of Hankel matrix
    _svd = numerics.svd_routine(**options.get("svd", {}))

    U,S,V = _svd(H0)
    SigmaInvSqrt = np.diag(S[:r]**-0.5)
    SigmaSqrt = np.diag(S[:r]**0.5)
    Ur = U[:,:r]
    Vr = V[:,:r]

    # get A from SVD and shifted Hankel matrix
    Ar = SigmaInvSqrt @ Ur.T.conj() @ H1 @ Vr @ SigmaInvSqrt

    # get B and C
    Br = (SigmaSqrt @ Vr.T.conj())[:,:q]
    Cr = (Ur @ SigmaSqrt)[:p,:]

    return (Ar,Br,Cr,Dr)


def era_dc(Y,**options):
    """
    System realization from Markov parameters (discrete impulse response data).
    Eigensystem Realization Algorithm with Data Correlations (ERA/DC) [3]_.


    :param Y:       Markov parameters. dimensions: :math:`(p,q,nt)`, where :math:`p` = number of outputs,
                    :math:`q` = number of inputs, and :math:`nt` = number of Markov parameters.
    :type Y:        array
    :param horizon: number of block rows in Hankel matrix = order of observability matrix.
                    default: :math:`min(150, (nt-1)/2)`
    :type horizon:  int, optional
    :param nc:      number of block columns in Hankel matrix = order of controllability matrix.
                    default: :math:`min(150, max(nt-1-` ``horizon``:math:`, (nt-1)/2))`
    :type nc:       int, optional
    :param order:   model order. default: :math:`min(20,` ``horizon``:math:`/2)`
    :type order:    int, optional
    :param a:       :math:`(\\alpha)` number of block rows in Hankel of correlation matrix. default: 0
    :type a:        int, optional
    :param b:       :math:`(\\beta)` number of block columns in Hankel of correlation matrix. default: 0
    :type b:        int, optional
    :param l:       initial lag for data correlations. default: 0
    :type l:        int, optional
    :param g:       lags (gap) between correlation matrices. default: 1
    :type g:        int, optional

    :return:        realization in the form of state space coefficients ``(A,B,C,D)``
    :rtype:         tuple of arrays

    References
    ----------
    .. [3]  Juang, J. N., Cooper, J. E., & Wright, J. R. (1987). An eigensystem realization algorithm
            using data correlations (ERA/DC) for modal parameter identification.
            (https://ntrs.nasa.gov/citations/19870035963)    
    """
    p,q,nt = Y.shape # p = number of outputs, q = number of inputs, nt = number of timesteps

    no = options.get("no",
         options.get("horizon",
                     None))
    nc = options.get("nc",
                     None)
    a = options.get("a", 0)
    b = options.get("b", 0)
    l = options.get("l", 0)
    g = options.get("g", 1)

    # get D from first p x q block of impulse response
    Dr = Y[:,:,0]  # first block of output data

    # size of Hankel matrix
    if no is None:
        no = min(150, int((nt-1)/2))
    if nc is None:
        nc = min(150, max(nt-1-no, int((nt-1)/2)))
    # make sure there are enough timesteps to assemble the Hankel matrices
    assert nt >= l+(a+1+b+1)*g+no+nc

    r = options.get("r", 
        options.get("order",
                    min(20, int(no/2))))

    # Hankel matrix of impulse response (Markov parameters)
    H = np.zeros((p*(no), q*(nc+l+(a+1+b+1)*g)))
    for i in range(no):
        for j in range(nc+l+(a+1+b+1)*g):
            H[p*i:p*(i+1), q*j:q*(j+1)] = Y[:,:,i+j+1]
    H0 = H[:,:q*nc]
    assert H0.shape == (p*(no), q*(nc))

    dimR = p*no # Dimension of square correlation matrices
    dimHRl = (dimR*(a+1), dimR*(b+1)) # Dimension of Hankel matrix of correlation matrices
    HRl = np.zeros(dimHRl) # Hankel matrix of correlation matrices
    HRl1 = np.zeros(dimHRl) # Shifted Hankel matrix of correlation matrices
    for i in range(a+1):
        for j in range(b+1):
            Hl = H[:, q*(l+1+(i+j)*g):q*(l+1+(i+j)*g+nc)]
            Hl1 = H[:, q*(l+1+(i+j)*g+1):q*(l+1+(i+j)*g+nc+1)]
            assert Hl.shape == Hl1.shape == (p*(no), q*(nc))
            R = Hl@H0.T       # correlation matrix
            R1 = Hl1@H0.T     # shifted correlation matrix
            assert R.shape == R1.shape == (dimR,dimR)
            HRl[dimR*i:dimR*(i+1), dimR*j:dimR*(j+1)] = R
            HRl1[dimR*i:dimR*(i+1), dimR*j:dimR*(j+1)] = R1

    # reduced SVD of Hankel matrix of correlation matrices
    _svd = numerics.svd_routine(**options.get("svd", {}))

    U,S,V = _svd(HRl)
    SigmaInvSqrt = np.diag(S[:r]**-0.5)
    SigmaSqrt = np.diag(S[:r]**0.5)
    Ur = U[:,:r]
    Vr = V[:,:r]

    # get A from SVD and shifted Hankel matrix of correlation matrices
    Ar = SigmaInvSqrt @ Ur.T.conj() @ HRl1 @ Vr @ SigmaInvSqrt

    # get B and C
    Br = ((SigmaInvSqrt @ Ur.T.conj())[:,:dimR] @ H0)[:,:q]
    Cr = (Ur @ SigmaSqrt)[:p,:]

    return (Ar,Br,Cr,Dr)


def _blk_3(i, CA, U):
    return i, np.einsum('kil,klj->ij', CA[:i,:,:], U[-i:,:,:])


def srim(inputs,outputs,**options):
    """
    System realization from input and output data, with output error minimization method.
    System Realization Using Information Matrix (SRIM) [4]_.
    
    :param inputs:  input time history. dimensions: :math:`(q,nt)`, where
                    :math:`q` = number of inputs, and :math:`nt` = number of timesteps
    :type inputs:   array
    :param outputs: output response history.
                    dimensions: :math:`(p,nt)`, where :math:`p` = number of outputs, and
                    :math:`nt` = number of timesteps
    :type outputs:  array
    :param horizon: number of steps used for identification (prediction horizon).
                    default: :math:`min(300, nt)`
    :type horizon:  int, optional
    :param order:   model order. default: :math:`min(20,` ``horizon``:math:`/2)`
    :type order:    int, optional
    :param full:    if True, full SVD. default: True
    :type full:     bool, optional
    :param find:    "ABCD" or "AC". default: "ABCD"
    :type find:     string, optional
    :param threads: number of threads used during the output error minimization method.
                    default: 6
    :type threads:  int, optional
    :param chunk:   chunk size in output error minimization method. default: 200
    :type chunk:    int, optional

    :return:        realization in the form of state space coefficients ``(A,B,C,D)``
    :rtype:         tuple of arrays

    References
    ----------
    .. [4]  Juang, J. N. (1997). System realization using information matrix. Journal
            of Guidance, Control, and Dynamics, 20(3), 492-500.
            (https://doi.org/10.2514/2.4068)
    """
    lsq_solve = numerics.lsq_solver(options.get("lsq", {}))

    if len(inputs.shape) == 1:
        inputs = inputs[None,:]
    if len(outputs.shape) == 1:
        outputs = outputs[None,:]

    if inputs.shape[0] > inputs.shape[1]:
        warnings.warn("input data has more channels (dim 1) than timesteps (dim 2)")
    if outputs.shape[0] > outputs.shape[1]:
        warnings.warn("output data has more channels (dim 1) than timesteps (dim 2)")

    q,nt = inputs.shape
    p = outputs.shape[0]
    assert nt == outputs.shape[1]

    no = options.get("no",
         options.get("horizon",
                     min(300, nt)))

    r = options.get("r", 
        options.get("order",
                    min(20, int(no/2))))

    full = options.get("full", True)
    find = options.get("find","ABCD")
    threads = options.get("threads",6)
    chunk = options.get("chunk", 200)

    # maximum possible number of columns in the Y and U data matrices
    ns = nt-1-no+2

    assert no >= r/p + 1    # make sure prediction horizon is large enough that
                            # observability matrix is full rank (Juang Eq. 8)

    Yno = np.zeros((p*no,ns))
    Uno = np.zeros((q*no,ns))

    # progress_bar = lambda arg, **kwds: (i for i in arg)

    # Construct Y (output) & U (input) data matrices (Eqs. 3.58 & 3.60 Arici 2006)
    for i in range(no):
        Yno[i*p:(i+1)*p,:] = outputs[:,i:ns+i]
        Uno[i*q:(i+1)*q,:] = inputs[:,i:ns+i]


    # 2b. Compute the correlation terms and the coefficient matrix 
    #     (Eqs. 3.68 & 3.69).

    # Compute the correlation terms (Eq. 3.68)
    Ryy = Yno@Yno.T/ns
    Ruu = Uno@Uno.T/ns
    Ruy = Uno@Yno.T/ns

    assert Ryy.shape[0] == Ryy.shape[1] == no*p
    assert Ruy.shape[0] == no*q
    assert Ruy.shape[1] == no*p

    # Compute the correlation matrix (Eq. 3.69)
    Rhh = Ryy - Ruy.T@lin_solve(Ruu,Ruy)

    # 2c. Obtain observability matrix using full or partial decomposition 
    #     (Eqs. 3.72 & 3.74).
    if full:
        # Full Decomposition Method
        un,*_ = np.linalg.svd(Rhh,0)           # Eq. 3.74
        Observability = un[:,:r]               # Eq. 3.72
        A = lsq_solve(Observability[:(no-1)*p,:], Observability[p:no*p,:])
        C = Observability[:p,:]
    else:
        # Partial Decomposition Method
        un,*_ = np.linalg.svd(Rhh[:,:(no-1)*p+1],0)
        Observability = un[:,:r]
        A = lsq_solve(Observability[:(no-1)*p,:], Observability[p:no*p,:])
        C = un[:p,:]

    # Computation of system matrices B & D
    # Output Error Minimization

    # Setting up the Phi matrix
    Phi  = np.zeros((p*ns, r+p*q+r*q))
    CA_powers = np.zeros((ns, p, A.shape[1]))
    CA_powers[0, :, :] = C
    A_p = A
    for pwr in range(1,ns):
        CA_powers[pwr,:,:] =  C@A_p
        A_p = A@A_p

    # First block column of Phi
    for df in range(ns):
        Phi[df*p:(df+1)*p,:r] = CA_powers[df,:,:]

    # Second block column of Phi
    Ipp = np.eye(p)
    for i in range(ns):
        Phi[i*p:(i+1)*p, r:r+p*q] = np.kron(inputs[:,i],Ipp)

    # Third block column of Phi
    In1n1 = np.eye(r)
    cc = r + p*q + 1
    dd = r + p*q + r*q

    krn = np.array([np.kron(inputs[:,i],In1n1) for i in range(ns)])

    # Execute a loop in parallel that looks something like:
    #    for i in  range(1,ns):
    #        Phi[] = _blk_3(i, CA_Powers, np.flip(...))

    if "b" not in find.lower() and "d" not in find.lower():
        return (A,None,C,None)

    with multiprocessing.Pool(threads) as pool:
        for i,res in progress_bar(
                pool.imap_unordered(
                    partial(_blk_3, CA=CA_powers,U=np.flip(krn,0)),
                    range(1,ns),
                    chunk
                ),
                total = ns
            ):
            Phi[i*p:(i+1)*p,cc-1:dd] = res

    y = outputs[:,:ns].flatten()

    teta = lsq_solve(Phi,y)

    x0 = teta[:r]
    dcol = teta[r:r+p*q]
    bcol = teta[r+p*q:r+p*q+r*q]

    D = np.zeros((p,q))
    B = np.zeros((r,q))
    for wq in range(q):
        D[:,wq] = dcol[wq*p:(wq+1)*p]

    for ww in range(q):
        B[:,ww] = bcol[ww*r:(ww+1)*r]

    assert A.shape[0] == A.shape[1] == r

    return A,B,C,D




import numpy as np
from numpy import pi

def EnergyCondensedEMAC(emac,phi):
    n,p = emac.shape
    assert emac.shape == phi.shape
    "Equation 3.93"
    return np.array([
        sum(emac[i,j]*abs(phi[i,j])**2 
            for j in range(p)
        )/np.real(phi[i,:].conjugate().transpose()@phi[i,:])
        for i in range(n) 
    ])

def EMAC_Matrix(Phi_final, Phi_final_hat):
    p,n = Phi_final.shape
    emac = np.zeros((n,p))
    for i in range(n):
        for j in range(p):
            # print(f"{Phi_final[j,i]=}, {Phi_final_hat[j,i]=}")
            Rij = min(
                (abs(Phi_final[j,i])/abs(Phi_final_hat[j,i])),
                (abs(Phi_final_hat[j,i])/abs(Phi_final[j,i]))
            )
            Pij = np.angle(Phi_final[j,i]/Phi_final_hat[j,i])
            Wij = max(
                1 - abs(Pij)/(pi/4),
                0
            )
            emac[i,j] = Rij*Wij
    return emac


def OutputEMAC(A,C,Psi=None,Gam=None,**options):
    """
    :param outlook:         number of timesteps for which to consider temporal consistency. default: 100
    :param Psi:             eigenvectors of A, as columns of a matrix. can be reused from modal.system_modes().
    :param Gam:             eigenvalues of A, as items of a vector. can be reused from modal.system_modes().
    :param Observability:   Observability matrix; can be reused from :func:`mdof.realize.srim`.
    """
    # """
    # :param p:               number of outputs
    # :param n:               model order (number of system variables)

    # Examples
    # --------
    # >>> a = np.random.randn(9, 6) + 1j*np.random.randn(9, 6)
    # >>> b = np.random.randn(2, 7, 8, 3) + 1j*np.random.randn(2, 7, 8, 3)

    # Reconstruction based on full SVD, 2D case:

    # >>> U, S, Vh = np.linalg.svd(a, full_matrices=True)
    # >>> U.shape, S.shape, Vh.shape
    # ((9, 9), (6,), (6, 6))
    # >>> np.allclose(a, np.dot(U[:, :6] * S, Vh))
    # True
    # >>> smat = np.zeros((9, 6), dtype=complex)
    # >>> smat[:6, :6] = np.diag(S)
    # >>> np.allclose(a, np.dot(U, np.dot(smat, Vh)))
    # True

    # Reconstruction based on reduced SVD, 2D case:

    # >>> U, S, Vh = np.linalg.svd(a, full_matrices=False)
    # >>> U.shape, S.shape, Vh.shape
    # ((9, 6), (6,), (6, 6))
    # >>> np.allclose(a, np.dot(U * S, Vh))
    # True
    # >>> smat = np.diag(S)
    # >>> np.allclose(a, np.dot(U, np.dot(smat, Vh)))
    # True

    # Reconstruction based on full SVD, 4D case:

    # >>> U, S, Vh = np.linalg.svd(b, full_matrices=True)
    # >>> U.shape, S.shape, Vh.shape
    # ((2, 7, 8, 8), (2, 7, 3), (2, 7, 3, 3))
    # >>> np.allclose(b, np.matmul(U[..., :3] * S[..., None, :], Vh))
    # True
    # >>> np.allclose(b, np.matmul(U[..., :3], S[..., None] * Vh))
    # True

    # Reconstruction based on reduced SVD, 4D case:

    # >>> U, S, Vh = np.linalg.svd(b, full_matrices=False)
    # >>> U.shape, S.shape, Vh.shape
    # ((2, 7, 8, 3), (2, 7, 3), (2, 7, 3, 3))
    # >>> np.allclose(b, np.matmul(U * S[..., None, :], Vh))
    # True
    # >>> np.allclose(b, np.matmul(U, S[..., None] * Vh))
    # True
    # """
    p,n = C.shape
    assert A.shape == (n,n)
    no = options.get("outlook",
         options.get("no",
         options.get("horizon",
                     100)))
    Observability = options.get("Observability",
                                None)
    """Output EMAC (Eqs. 3.88-3.89)"""
    if Gam is None:
        assert Psi is None
        from mdof.modal import condeig 
        Psi,Gam,_ = condeig(A)
    if Observability is None:
        Observability = np.empty((no,p,n))
        Observability[0,:,:] = C
        A_pwr = A
        for pwr in range(1,no):
            Observability[pwr,:,:] =  C@A_pwr
            A_pwr = A@A_pwr
        Observability = Observability.reshape((no*p,n))
    Phi_final = Observability[-p:,:]@Psi                    # identified modal observability at the last timestep (last block row)
    Phi_final_hat = C@Psi@np.diag(Gam**(no-1))              # expected modal observability at the last timestep
    assert Phi_final.shape == Phi_final_hat.shape == (p,n)
    emaco  = EMAC_Matrix(Phi_final,Phi_final_hat)
    return EnergyCondensedEMAC(emaco,(C@Psi).T)             # DIM: nxp, nxp

def MPC(A,C,Psi=None):
    """a) Modal Phase Collinearity (MPC) [Eqs. 3.85-3.87]"""
    if Psi is None:
        from mdof.modal import condeig 
        Psi,_,_ = condeig(A)
    _,n = C.shape
    assert Psi.shape == (n,n)
    modes_raw = C@Psi
    s11, s22, s12 = np.zeros((3,n))
    nu, mpc = np.zeros((2,n))
    lam = np.zeros((2,n))
    for i in range(n):
        mode_i = modes_raw[:,i]
        s11[i] = np.real(mode_i).conjugate().transpose()@np.real(mode_i)
        s22[i] = np.imag(mode_i).conjugate().transpose()@np.imag(mode_i)
        s12[i] = np.real(mode_i).conjugate().transpose()@np.imag(mode_i)
        # print(f"{mode_i=}, {s11[i]=}, {s22[i]=}, {s12[i]=}")
        if s12[i] == 0:
            nu[i] = np.nan
        else:
            nu[i]= (s22[i]-s11[i])/(2*s12[i])
        lam[0,i] = (s11[i]+s22[i])/2 + s12[i]*np.sqrt(nu[i]**2+1)
        lam[1,i] = (s11[i]+s22[i])/2 - s12[i]*np.sqrt(nu[i]**2+1)
        mpc[i]   = ((lam[0,i]-lam[1,i])/(lam[0,i]+lam[1,i]))**2
    return mpc


import numpy as np
import warnings

# inputs = input data. dimensions of inputs: q x nt, where nt = number of timesteps.
# outputs = response data due to input data. dimensions of outputs: p x nt.
# m = number of Markov parameters (impulse response timesteps), not including timestep zero, to solve for.
def okid(inputs,outputs,**options):
    """
    Identify Markov parameters, or discrete impulse response data, for a given set of input and output data.
    Observer Kalman Identification Algorithm (OKID) [1]_.

    :param inputs:  input time history. dimensions: :math:`(q,nt)`, where
                    :math:`q` = number of inputs, and :math:`nt` = number of timesteps
    :type inputs:   array
    :param outputs: output response history.
                    dimensions: :math:`(p,nt)`, where :math:`p` = number of outputs, and
                    :math:`nt` = number of timesteps
    :type outputs:  array
    :param m:       number of Markov parameters to compute. default: :math:`min(300, nt)`
    :type m:        int, optional

    :return: the Markov parameters, with dimensions :math:`(p,q,m+1)`
    :rtype: array

    References
    ----------
    .. [1]  Juang, J. N., Phan, M., Horta, L. G., & Longman, R. W. (1993). Identification of
            observer/Kalman filter Markov parameters-Theory and experiments. Journal of Guidance,
            Control, and Dynamics, 16(2), 320-329. (https://doi.org/10.2514/3.21006)
    """

    if len(inputs.shape) == 1:
        inputs = inputs[None,:]
    if len(outputs.shape) == 1:
        outputs = outputs[None,:]
    
    if inputs.shape[0] > inputs.shape[1]:
        warnings.warn("input data has more channels (dim 1) than timesteps (dim 2)")
    if outputs.shape[0] > outputs.shape[1]:
        warnings.warn("output data has more channels (dim 1) than timesteps (dim 2)")

    q,nt = inputs.shape
    p = outputs.shape[0]
    assert nt == outputs.shape[1]
    
    m = options.get("m", None)
    if m is None:
        m = min(1000,nt)

    # Form data matrix V
    V = np.zeros((m+1,nt,q+p))                    # m+1 block rows, each with nt columns and q+p rows each.  DIM: (m+1)x(nt)x(q+p)
    for i in range(m+1):                          # From block row 0 to block row m
        V[i,-(nt-i):,:q] = inputs[:,:nt-i].T      # Populate the first q rows of the ith block row, last nt-i columns, with the first nt-i inputs.  DIM: (nt-i)x(q)
        V[i,-(nt-i):,-p:] = outputs[:,:nt-i].T    # Populate the last p rows of the ith block row, last nt-i columns, with the first nt-i outputs.  DIM: (nt-i)x(p)
    V = V.transpose((1,0,2)).reshape((nt,(m+1)*(p+q))).T    # Transpose and reshape data matrix to stack the block rows in.  DIM: ((m+1)(q+p))x(nt)
    # V = np.concatenate((V[:q,:],V[q+p:,:]))     # Remove rows q+1 to q+p (TODO: see if it's necessary to remove rows, or if solving for Ybar can include these)
    V = np.delete(V, slice(q,q+p), axis=0)        # Remove rows q+1 to q+p (TODO: see if it's necessary to remove rows, or if solving for Ybar can include these)

    # Solve for observer Markov parameters Ybar
    Ybar = outputs @ np.linalg.pinv(V,rcond=10e-3)  # DIM: (p)x(q+m(q+p))
    assert Ybar.shape == (p,q+m*(q+p))
    
    # Isolate system Markov parameters
    H = np.zeros((p,q,m+1))
    D = Ybar[:,:q] # feed-through term (or D matrix) is the first block column
    H[:,:,0] = D

    Y = np.zeros((p,q,m))
    Ybar1 = np.zeros((p,q,m))
    Ybar2 = np.zeros((p,p,m))
    
    for i in range(m):
        Ybar1[:,:,i] = Ybar[:,q+(q+p)*i : q+(q+p)*i+q]
        Ybar2[:,:,i] = Ybar[:,q+(q+p)*i+q : q+(q+p)*(i+1)]
    
    Y[:,:,0] = Ybar1[:,:,0] + Ybar2[:,:,0] @ D
    for k in range(1,m):
        Y[:,:,k] = Ybar1[:,:,k] + Ybar2[:,:,k] @ D
        for i in range(k-1):
            Y[:,:,k] += Ybar2[:,:,i] @ Y[:,:,k-i-1]

    H[:,:,1:] = Y

    return H



def lsq_solver(options):
    import numpy.linalg
    return lambda *args: numpy.linalg.lstsq(*args, rcond=None)[0]

def lin_solver():
    pass

def svd_routine(lapack_driver="gesvd", **kwds):
    import scipy
    # default lapack_driver="gesdd"
    def _svd(*args):
        U,S,V = scipy.linalg.svd(*args, lapack_driver=lapack_driver, **kwds)
        return U,S,V.T.conj()
    return _svd

def form_observability():
    pass

def form_controllability():
    pass

def decimate(series, decimation):
    import numpy as np
    if isinstance(series, np.ndarray):
        if len(series.shape) == 1:
            return series[np.arange(0,len(series),decimation)]
        else:
            return series[:,np.arange(0,series.shape[1],decimation)]
    if isinstance(series, list):
        return np.asarry(series)[np.arange(0,len(series),decimation)]
    
def block_hankel(series):

    pass


from typing import Any


class Config(dict):
    def __setattr__(self, name, value):
        self[name]  = value
        
    def __getattribute__(self, __name: str) -> Any:
        if __name in self:
            return self[__name]
        return super().__getattribute__(__name)


def _decimate(series, d):
    import numpy as np

    return series[np.arange(0,len(series),d)]

def extract_channels(event, channels, decimate=1):
    import numpy as np
    import sys

    def find(chan):
        #event.match("r", file_name=f".*{chan}\.[vV]2")
        return event.match("l", station_channel=f"{chan}")

    data = np.asarray([
        _decimate(find(chan).accel.data, d=decimate)
        for chan in channels if find(chan) is not None
    ])


    if len(data) == 0:
        raise ValueError("No channels found")
    
    elif len(data) != len(channels):
        print(f"Only extracted {len(data)} channels, {len(channels)-len(data)} missing.", file=sys.stderr)

    dt = find(channels[0]).accel["time_step"]*decimate
    return data, dt


def list_files(directory, pattern):
    from pathlib import Path
    return Path(directory).glob(pattern)


import mdof
from mdof import modal
from time import time
from control import ss, forced_response

def test_method(method, inputs, outputs, dt, t, **conf):
    time0 = time()
    A,B,C,D = mdof.system(method=method, inputs=inputs, outputs=outputs, **conf)
    time1 = time()
    modedict = modal.system_modes((A,B,C,D),dt,**conf)
    model = {
                "time":    time1-time0,
                "ypred":   forced_response(ss(A,B,C,D,dt), T=t, U=inputs, squeeze=False, return_x=False).outputs,
                "modes":   modedict,
                "period":  [1/value['freq'] for value in modedict.values()],
                "damping": [value['damp'] for value in modedict.values()]
            }
    return model

def mode_statistics(mode_results, key):
    import numpy as np
    values = [result[key] for results in mode_results for result in results]
    mean = np.mean(values)
    std = np.std(values)
    closest_values = [results[np.argmin(np.abs(mean-[result[key] for result in results]))] for results in mode_results]
    return [
        dict(**item, distance=(item[key]-mean)/std) for item in closest_values
    ]

def mode_set(mode_results, key):
    return 


# default imports

from .system import system
from .modal import system_modes, _condeig

# from control.matlab import impulse as _impulse
# def impulse(system, t, **kwds):
#     from control import ss
#     dt = t[1] - t[0]
#     a,t = _impulse(ss(*system, dt), t, **kwds)
#     return a.squeeze()*dt,t

def sysid(inputs, outputs, **options):
    """
    State space system realization from input and output data.

    :param inputs:      input time history. dimensions: :math:`(q,nt)`, where
                        :math:`q` = number of inputs, and :math:`nt` = number of timesteps
    :type inputs:       array
    :param outputs:     output response history.
                        dimensions: :math:`(p,nt)`, where :math:`p` = number of outputs, and
                        :math:`nt` = number of timesteps
    :type outputs:      array
    :param method:      system identification method. default is "srim", other options are "okid-era" and "okid-era-dc".
    :type method:       string, optional
    :param decimation:  decimation factor. default: 8
    :type decimation:   int, optional

    :return:            system realization in the form of state space coefficients ``(A,B,C,D)``
    :rtype:             tuple of arrays
    """
    return system(inputs, outputs, **options)

def eigid(inputs, outputs, **options):
    """
    System eigenspace identification from input and output data.

    :param inputs:      input time history. dimensions: :math:`(q,nt)`, where
                        :math:`q` = number of inputs, and :math:`nt` = number of timesteps
    :type inputs:       array
    :param outputs:     output response history.
                        dimensions: :math:`(p,nt)`, where :math:`p` = number of outputs, and
                        :math:`nt` = number of timesteps
    :type outputs:      array
    :param dt:          timestep.
    :type dt:           float
    :param decimation:  decimation factor. default: 1
    :type decimation:   int, optional

    :return:            `(eigenvalues, eigenvectors)`
    :rtype:             tuple of 1D array, ND array
    """
    A,_,_,_ = sysid(inputs, outputs, **options)
    decimation = options.get("decimation", 1)
    # eigendecomp A
    Psi,Gam,_ = _condeig(A)  # eigenvectors (Psi) & eigenvalues (Gam) of the matrix A
    return (Gam,Psi)

def modes(inputs, outputs, dt, **options):
    """
    Modal identification using state space system realization from input and output data.

    :param inputs:      input time history. dimensions: :math:`(q,nt)`, where
                        :math:`q` = number of inputs, and :math:`nt` = number of timesteps
    :type inputs:       array
    :param outputs:     output response history.
                        dimensions: :math:`(p,nt)`, where :math:`p` = number of outputs, and
                        :math:`nt` = number of timesteps
    :type outputs:      array
    :param dt:          timestep.
    :type dt:           float
    :param method:      system identification method. default is "srim", other options are
                        "okid-era" and "okid-era-dc".
    :type method:       string, optional
    :param decimation:  decimation factor. default: 8
    :type decimation:   int, optional

    :return:            system modes, including natural frequencies, damping ratios, mode shapes,
                        condition numbers, and modal validation metrics EMAC and MPC.
    :rtype:             dictionary
    """
    realization = system(inputs, outputs, **options)

    return modal.system_modes(realization, dt)




from mdof import modes
from matplotlib import pyplot as plt
import numpy as np

def stabilization(inputs, outputs, dt, plotly=False, **options):
    """
    Stabilization diagram of modal properties from state space system identification, based on model order.

    :param inputs:      input time history. dimensions: :math:`(q,nt)`, where
                        :math:`q` = number of inputs, and :math:`nt` = number of timesteps
    :type inputs:       array
    :param outputs:     output response history.
                        dimensions: :math:`(p,nt)`, where :math:`p` = number of outputs, and
                        :math:`nt` = number of timesteps
    :type outputs:      array
    :param orders:      range of system orders. default: (2,100,2)
    :type orders:       tuple (`start`, `stop`, `step`), optional
    :param method:      system identification method. default is "srim", other options are "okid-era" and "okid-era-dc".
    :type method:       string, optional
    :param decimation:  decimation factor. default: 1
    :type decimation:   int, optional

    :return:            system realization in the form of state space coefficients ``(A,B,C,D)``
    :rtype:             tuple of arrays
    """
    orders = options.get("orders", (2,100,2)) 
    mode_properties = np.array([[order, 1/mode["freq"], mode["damp"]] for order in range(*orders) for mode in modes(inputs, outputs, dt, order=order, **options).values()])
    if plotly:
        import plotly.graph_objects as go
        from plotly.subplots import make_subplots
        fig = make_subplots(rows=1, cols=2, shared_yaxes=True, y_title="Model Order")
        fig.add_trace(go.Scatter(x=mode_properties[:,1],y=mode_properties[:,0], mode="markers", marker=dict(size=8)),
                      row=1, col=1)
        fig.add_trace(go.Scatter(x=mode_properties[:,2],y=mode_properties[:,0], mode="markers", marker=dict(size=8)),
                      row=1, col=2)
        fig.update_layout(height=300, margin=dict(l=70, r=20, t=20, b=20), showlegend=False)
        fig.layout.xaxis.title="Period (s)"
        fig.layout.xaxis2.title="Damping Ratio"
        fig.show(renderer="notebook_connected")
    else:
        fig, ax = plt.subplots(2,1,figsize=(6,6))
        ax[0].plot(mode_properties[:,1],mode_properties[:,0],'o',markersize=4)
        ax[0].set_title("Periods")
        ax[0].set_ylabel("Model Order")
        ax[1].plot(mode_properties[:,2],mode_properties[:,0],'o',markersize=4)
        ax[1].set_title("Damping Ratios")
        ax[1].set_ylabel("Model Order")





# gets called when running mdof from the command line:
#    python -m mdof ...
#
import ast
import sys
import json

import mdof
import mdof.modal
import quakeio
import numpy as np
# from .okid import parse_okid

HELP = """
mdof [-p|w <>...] <method> <event> <inputs> <outputs>
mdof [-p|w <>...] <method> -i <inputs> -o <outputs>

-i/--inputs  FILE...   Input data
-o/--outputs FILE...   Output data

Methods:
  <method> <outputs>  <plots>
    srim    {dftmcs}   {m}
    okid    {dftmcs}   {m}
    spec    {ft}       {a}
    four    {ft}       {a}
    test

Outputs options
-p/--plot
    a/--accel-spect
-w/--write
    ABCD   system matrices
    d      damping
    freq   frequency
    cycl   cyclic frequency
    t      period
    m      modes
    c      condition-number
"""

class JSON_Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        return json.JSONEncoder.default(self, obj)


# def _decimate(series, decimation, **options):
#     import numpy as np
#     return series[np.arange(0,len(series),decimation)]


# def extract_channels(event, channels, decimate=1, permissive=True):
def extract_channels(event, channels, permissive=True):
    import numpy as np
    import sys

    def find(chan):
        #event.match("r", file_name=f".*{chan}\.[vV]2")
        return event.match("l", station_channel=f"{chan}")

    data = np.asarray([
        # _decimate(find(chan).accel.data, d=decimate)
        find(chan).accel.data
        for chan in channels if find(chan) is not None
    ])

    if len(data) == 0:
        raise ValueError(f"No channels found, requested '{channels}'")

    elif len(data) != len(channels):
        if permissive:
            print(f"Only extracted {len(data)} channels, {len(channels)-len(data)} missing.", file=sys.stderr)
        else:
            raise ValueError("Could not extract all desired channels")

    # dt = find(channels["inputs"]).accel["time_step"]*decimate
    dt = find(channels[0]).accel["time_step"]
    return data, dt



def parse_time(argi, config, channels, method=None):
    help = f"""\
    mdof {method} <event>


    --inputs  <int>
    --outputs <int>

    Options
    --threads <int>
    """
    damping  = None
    for arg in argi:
        if arg == "--threads" and method == "response":
            config["threads"] = int(next(argi))

        elif arg == "--config":
            conf_arg = json.loads(next(argi))
            for i in conf_arg.pop("channels", []):
                channels[i["type"]+"s"].append(i["id"])
            for i in {"inputs", "outputs"}:
                if i in conf_arg:
                    channels[i] = conf_arg.pop(i)
            config.update(conf_arg)

        elif arg == "--inputs":
            channels["intpus"] = next(argi)

        elif arg == "--outputs":
            channels["outputs"] = next(argi)

        elif arg == "--decimate":
            config["decimate"] = int(next(argi))

        elif arg == "--damping" and method == "response":
            damp = next(argi)
            try:
                config["damping"] = [float(damp)]
            except:
                config["damping"] = ast.literal_eval(damp)

        elif arg == "--periodband":
            config["period_band"] = tuple(float(x) for x in (next(argi).split(" ")))

        elif arg == "-h" or arg == "--help":
            print(help)
            sys.exit()

        else:
            config["event_file"] = arg

    event = quakeio.read(config["event_file"], exclusions=["filter*"])

    try:
        # inputs,  dt = extract_channels(event, [channels["inputs"]], decimate=decimate)
        inputs,  dt = extract_channels(event, [channels["inputs"]])
        # outputs, dt = extract_channels(event, [channels["outputs"]], decimate=decimate)
        outputs, dt = extract_channels(event, [channels["outputs"]])
    except Exception as e:
        print(json.dumps({"error": str(e), "data": []}))
        return

    import mdof.transform
    from mdof.modal import spectrum_modes
    f = {
        "response": mdof.transform.response_transfer,
        "fourier":  mdof.transform.fourier_transfer,
    }[method]

    periods, amplitudes = spectrum_modes(
                          *f(inputs=inputs.flatten(), outputs=outputs.flatten(), step = dt, **config)
                        )
    output = [
            {"period": period, "amplitude": amplitude}
            for period, amplitude in zip(periods, amplitudes)
    ]
    print(json.dumps({"data": output}, cls=JSON_Encoder, indent=4))


def parse_srim(argi, config, channels, method=None):
    help="""
    SRIM -- System Identification with Information Matrix

    Parameters
    no          order of the observer Kalman ARX filter (formerly p).
    r           size of the state-space model used for
                representing the system. (formerly orm/n)
    """
#   config.update({"p"  :  5, "orm":  4})

    for arg in argi:
        if arg == "--arx-order":
            config["no"] = int(next(argi))

        elif arg == "--config":
            conf_arg = json.loads(next(argi))
            for i in conf_arg.pop("channels", []):
                channels[i["type"]+"s"].append(i["id"])
            for i in {"inputs", "outputs"}:
                if i in conf_arg:
                    channels[i] = conf_arg.pop(i)
            config.update(conf_arg)

        elif arg == "--dt":
            config["dt"] = float(next(argi))

        elif arg == "--ss-size":
            config["r"] = int(next(argi))

        elif arg in ["--help", "-h"]:
            print(help)
            sys.exit()

        elif arg == "--inputs":
            channels["inputs"] = json.loads(next(argi))

        elif arg == "--outputs":
            channels["outputs"] = json.loads(next(argi))

        elif arg == "--decimate":
            config["decimate"] = int(next(argi))

        elif arg == "--":
            continue

        else:
            config["event_file"] = arg

    event = quakeio.read(config["event_file"], exclusions=["filter*"])
    try:
        # inputs,  dt = extract_channels(event, [channels["inputs"]], decimate=decimate)
        inputs,  dt = extract_channels(event, channels["inputs"])
        # outputs, dt = extract_channels(event, [channels["outputs"]], decimate=decimate)
        outputs, dt = extract_channels(event, channels["outputs"])
    except Exception as e:
        print(json.dumps({"error": str(e), "data": []}))
        return

    config["dt"] = dt

    try:
        realization = mdof.system(inputs=inputs, outputs=outputs, full=True, method=method, **config)
        ss_modes = mdof.modal.system_modes(realization,dt,decimation=config.get("decimate",1))

    except Exception as e:
        print(json.dumps({"error": str(e), "data": []}))
        return

    output = [
        {
            "period":  1/mode["freq"],
            "frequency": mode["freq"],
            "damping":   mode["damp"],
            "shape":     np.real(mode["modeshape"]),
            "emac":      mode["energy_condensed_emaco"],
            "mpc":       mode["mpc"],
        }
        for mode in sorted(ss_modes.values(), key=lambda item: item["freq"])
    ]

    print(json.dumps({"data": output}, cls=JSON_Encoder, indent=4))
    return config


def parse_args(args):
    outputs = []
    sub_parsers = {
        "srim":     parse_srim,
        "response": parse_time,
        "fourier":  parse_time,
        "test": parse_srim,
        "okid": parse_srim,
        "okid": parse_srim
    }
    method = None
    config = {}
    channels = {
#           "operation": None,
#           "protocol": None,
            "inputs":  [],
            "outputs": []
    }

    argi = iter(args[1:])
    for arg in argi:
        if arg == "-p":
            outputs.append(next(arg))

        elif arg == "--config":
            conf_arg = json.loads(next(argi))
            method = conf_arg.pop("method", method)
            for i in conf_arg.pop("channels", []):
                channels[i["type"]+"s"].append(i["id"])
            config.update(conf_arg)

        elif arg in ["--help", "-h"]:
            print(HELP)
            sys.exit()

#       elif arg == "--protocol":
#           config["protocol"] = next(arg)

        # Positional args
        elif method is None:
            method = arg
            break

        elif arg == "--":
            continue

        else:
            print(HELP)
            sys.exit()


    return sub_parsers[method](argi, config, channels, method=method), outputs


def main():
    parse_args(sys.argv)

if __name__ == "__main__":
    main()
    sys.exit()




import numpy as np
from matplotlib import pyplot as plt
import scienceplots
import plotly.graph_objects as go

plt.style.use(["poster"])# ,"science", "notebook"])

nln = "\n"

DEFAULT_PLOTLY_COLORS=[
    '#1f77b4',  # muted blue
    '#ff7f0e',  # safety orange
    '#2ca02c',  # cooked asparagus green
    '#d62728',  # brick red
    '#9467bd',  # muted purple
    '#8c564b',  # chestnut brown
    '#e377c2',  # raspberry yogurt pink
    '#7f7f7f',  # middle gray
    '#bcbd22',  # curry yellow-green
    '#17becf'   # blue-teal
]

def print_modes(modes, Tn=None, zeta=None):

    header = "       T(s)        \N{Greek Small Letter Zeta}        EMACO      MPC       EMACO*MPC"

    if Tn is not None:
        header += "     T % error"

    if zeta is not None:
        header += "    \N{Greek Small Letter Zeta} % error"

    print("Spectral quantities:")
    print(f"{header}")
    for mode in sorted(modes.values(), key=lambda x: x["freq"]):
        f = mode["freq"]
        z = mode["damp"]
        emaco = mode["energy_condensed_emaco"]
        mpc = mode["mpc"]
        row = f"      {1/f: <9.4}  {z: <9.4}  {emaco: <9.4}  {mpc: <9.4}  {emaco*mpc: <9.4}"
        if Tn is not None:
            row += f"    {100*(1/f-Tn)/(Tn): <9.4}"
        if zeta is not None:
            row += f"    {100*(z-zeta)/zeta: <9.4}"
        print(row)
    print("Mean Period(s):", np.mean([1/v["freq"] for v in modes.values()]))
    print("Standard Dev(s):", np.std([1/v["freq"] for v in modes.values()]))

def plot_models(models, Tn, zeta):
    fig, ax = plt.subplots(2, 3, constrained_layout=True, sharex=True, figsize=(13,6.5))

    period = [models[method]["period"][0] for method in models]
    ax[0,0].bar(["true"]+list(models), [Tn]+period)
    ax[0,0].set_title("Periods")# , fontsize=14)
    ax[0,0].set_ylabel("Period (s)")# , fontsize=13)

    period_errors = [100*(p-Tn)/Tn for p in period]
    ax[1,0].bar(models.keys(), period_errors, color=None, edgecolor="k", linewidth=0.5)
    ax[1,0].set_title("Period Errors")# , fontsize=14)
    ax[1,0].set_ylabel("Percent Error (%)")# , fontsize=13)
    ax[1,0].set_xlabel("Method")# , fontsize=13)

    damp = [models[method]["damping"][0] for method in models]
    ax[0,1].bar(["true"]+list(models), [zeta]+damp)
    ax[0,1].set_title("Damping")# , fontsize=14)
    ax[0,1].set_ylabel("Damping Ratio")# , fontsize=13)

    damping_errors = [100*(d-zeta)/zeta for d in damp]
    ax[1,1].bar(models.keys(), damping_errors, color=None, edgecolor="k", linewidth=0.5)
    ax[1,1].set_title("Damping Errors")# , fontsize=14)
    ax[1,1].set_ylabel("Percent Error (%)")# , fontsize=13)
    ax[1,1].set_xlabel("Method")# , fontsize=13)

    ax[0,2].axis('off')

    times_list = [models[method]["time"] for method in models]
    ax[1,2].bar(models.keys(), times_list, color=None, edgecolor="k", linewidth=0.5)
    ax[1,2].set_title("Runtime")# , fontsize=14)
    ax[1,2].set_ylabel("time (s)")# , fontsize=13)
    ax[1,2].set_xlabel("Method")# , fontsize=13)

    for axi in fig.axes:
        axi.tick_params(axis='x', rotation=45)#, labelsize=12)
        # axi.tick_params(axis='y', labelsize=12)

    for i,error in zip([0,1,2],[period_errors,damping_errors,times_list]):
        rects = ax[1,i].patches
        for rect, label in zip(rects, error):
            label = f"{label: <5.3}"
            height = rect.get_height()
            ax[1,i].text(
                rect.get_x() + rect.get_width() / 2, height, label, ha="center", va="bottom"
            )
    
    fig.suptitle("Spectral Quantity Prediction with System Identification",fontsize=16)

def plot_io(inputs, outputs, t, title=None):
    fig, ax = plt.subplots(1,2,figsize=(8,3),constrained_layout=True)
    # fig, ax = plt.subplots(1,2,figsize=(8,3))
    color_cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
    if len(inputs.shape) > 1:
        for i in range(inputs.shape[0]):
            ax[0].plot(t,inputs[i,:])
    else:
        ax[0].plot(t,inputs)
    ax[0].set_xlabel("time (s)")# , fontsize=13)
    ax[0].set_ylabel("inputs")# , fontsize=13)
    if len(outputs.shape) > 1:
        for i in range(outputs.shape[0]):
            ax[1].plot(t,outputs[i,:], color=color_cycle[0], alpha=0.5)
    else:
        ax[1].plot(t,outputs)
    ax[1].set_xlabel("time (s)")# , fontsize=13)
    ax[1].set_ylabel("outputs")# , fontsize=13)
    fig.suptitle(title, fontsize=16)

def plot_pred(ytrue, models, t, title=None):
    color_cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
    fig, ax = plt.subplots(figsize=(8,4))
    if len(ytrue.shape) > 1:
        for i in range(ytrue.shape[0]):
            ax.plot(t,ytrue[i,:],label=["true" if i==0 else None][0],color=color_cycle[0],alpha=0.5)
    else:
        ax.plot(t,ytrue,label="true")
    if type(models) is np.ndarray:
        if len(models.shape) > 1:
            for i in range(models.shape[0]):
                ax.plot(t,models[i,:],"--",label=["prediction" if i==0 else None][0],color=color_cycle[1],alpha=0.5)
        else:
            ax.plot(t,models,"--",label=f"prediction")
    else:
        for method in models:
            if len(models[method]["ypred"].shape) > 1:
                for i in range(models[method]["ypred"].shape[0]):
                    ax.plot(t,models[method]["ypred"][i,:],"--",label=method)
            else:
                ax.plot(t,models[method]["ypred"],"--",label=method)
    ax.set_xlabel("time (s)")# , fontsize=13)
    ax.set_ylabel("outputs")# , fontsize=13)
    fig.legend(fontsize=12, frameon=True, framealpha=1)    
    ax.set_title(title) #, fontsize=14)

def plot_transfer(models, title=None, labels=None, plotly=False):
    if plotly:
        import plotly.graph_objects as go
        layout = go.Layout(
            title=title,
            xaxis=dict(
                title="Period (s)"
            ),
            yaxis=dict(
                title="Amplitude"
            ),
            width=600, height=300,
            margin=dict(l=70, r=20, t=20, b=20))
        fig = go.Figure(layout=layout)
        if type(models) is np.ndarray:
            if len(models.shape) > 2:
                for i in range(models.shape[0]):
                    fig.add_trace(go.Scatter(x=models[i,0],y=models[i,1]/max(models[i,1]),name=labels[i]))
            else:
                fig.add_trace(go.Scatter(x=models[0],y=models[1]/max(models[1]),name=labels))
        else:
            for method in models:
                fig.add_trace(go.Scatter(x=models[method][0],y=models[method][1]/max(models[method][1]),name=method))
        fig.show(renderer="notebook_connected")
    else:
        fig, ax = plt.subplots(figsize=(6,3))
        if type(models) is np.ndarray:
            if len(models.shape) > 2:
                for i in range(models.shape[0]):
                    ax.plot(models[i,0],models[i,1]/max(models[i,1]),label=labels[i])
            else:
                ax.plot(models[0],models[1]/max(models[1]),label=labels)
        else:
            for method in models:
                ax.plot(models[method][0],models[method][1]/max(models[method][1]),label=method)
        ax.set_xlabel("Period (s)")
        ax.set_ylabel("Amplitude")
        if labels is not None:
            ax.legend()#fontsize=12)
        ax.set_title(title)#, fontsize=14)

class FrequencyContent:
    def __init__(self, scale, period, xlabel, ylabel) -> None:
        self.scale = scale
        self.period = period
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.num_traces = 0
        layout = go.Layout(
            title=None,
            xaxis=dict(
                title=self.xlabel
            ),
            yaxis=dict(
                title=self.ylabel
            ),
            width=400, height=300,
            margin=dict(l=70, r=20, t=20, b=20))
        self.fig = go.Figure(layout=layout)

    def add(self, periods, amplitudes=None, label=None):
        if self.period:
            x_data = periods
        else:
            x_data = 1/periods
        if amplitudes is not None:
            if self.scale:
                y_data = amplitudes/max(amplitudes)
            else:
                y_data = amplitudes
            self.fig.add_trace(go.Scatter(x=x_data,y=y_data,name=label,showlegend=True,line_color=DEFAULT_PLOTLY_COLORS[self.num_traces%len(DEFAULT_PLOTLY_COLORS)]))
        else:
            N = len(x_data)
            x_data_vlines = np.array([x_data,x_data,np.full(N,None)]).transpose().reshape(-1)
            y_data_vlines = np.array([np.zeros(N),np.ones(N),np.full(N,None)]).transpose().reshape(-1)
            self.fig.add_trace(go.Scatter(x=x_data_vlines,y=y_data_vlines,mode='lines',name=label,showlegend=True,line_dash="dash",line_color=DEFAULT_PLOTLY_COLORS[self.num_traces%len(DEFAULT_PLOTLY_COLORS)]))
        self.num_traces += 1        

    def get_figure(self):
        return self.fig

def make_hover_data(data, ln=None):
    import numpy as np
    if ln is None:
        items = np.array([d.values for d in data])
        keys = data[0].keys()
    else:
        items = np.array([list(data.values())]*ln)
        keys = data.keys()
    return {
        "hovertemplate": "<br>".join(f"{k}: %{{customdata[{v}]}}" for v,k in enumerate(keys)),
        "customdata": list(items),
    }

if __name__ == "__main__":

    import numpy as np
    periods = np.array([0.1,0.3,0.5,0.7])
    amplitudes = np.array([0.1,0.2,1,0.2])

    plot = FrequencyContent(scale=True, period=True, xlabel="Period (s)", ylabel="Amplitude")

    plot.add(periods, amplitudes, label="R1")
    plot.add(periods[:2], label="SRIM")
    plot.add(periods[2:], label="SRIM")
    plot.add(periods, amplitudes-0.05, label="R1")


    fig:go.Figure = plot.get_figure()
    print(fig.to_json())
